"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.checkDatasetMatchesFormat = exports.deriveDatasetFormat = exports.getAllFilesInFolder = exports.DatasetConverterHelperCli = void 0;
const path_1 = __importDefault(require("path"));
const fs_1 = __importDefault(require("fs"));
const papaparse_1 = __importDefault(require("papaparse"));
const inquirer_1 = __importDefault(require("inquirer"));
const label_file_types_1 = require("../shared/uploader/annotations-parsing-shared/label-file-types");
const parse_label_json_1 = require("../shared/uploader/annotations-parsing-shared/parse-label-json");
const fs_helpers_1 = require("./fs-helpers");
const parse_label_csv_1 = require("../shared/uploader/annotations-parsing-shared/parse-label-csv");
const bounding_box_file_types_1 = require("../shared/bounding-box-file-types");
const label_files_shared_1 = require("../shared/uploader/annotations-parsing-shared/label-files-shared");
const dataset_converter_1 = require("../shared/uploader/annotations-parsing-shared/dataset-converter");
const xmldom_1 = require("@xmldom/xmldom");
const parse_label_xml_1 = require("../shared/uploader/annotations-parsing-shared/parse-label-xml");
const image_size_1 = __importDefault(require("image-size"));
const js_yaml_1 = __importDefault(require("js-yaml"));
/**
 * This class implements a range of helper functions to convert dataset annotations into a format usable in EI.
 * This class works specifically in the CLI.
 */
class DatasetConverterHelperCli extends dataset_converter_1.DatasetConverterHelper {
    /**
     * Create a new dataset converter helper
     * @param format Dataset format
     * @param opts Additional options
     */
    constructor(format, opts) {
        super(format);
        this._opts = opts;
        this._successCount = 0;
        this._labelMapFiles = [];
    }
    /**
     * Convert annotations within a set of files
     * @param directory Directory to analyze
     */
    async convertDataset(files) {
        if (!this._opts.silent) {
            console.log('\nPre-processing annotation files...\n');
        }
        this.clearAnnotations();
        this._successCount = 0;
        // Store any actual samples to be uploaded
        this._samples = [];
        // Read all files in the directory, parsing any annotation files and listing any samples
        for (const file of files) {
            this.parseFile(file.filename, file.path);
        }
        // Get additional info for more complex formats
        if (this._datasetFormat.additionalOpts) {
            this._labelmap = await this.getLabelmap();
        }
        if (!this._opts.silent) {
            console.log(`\nDone. Annotation files parsed successfully: ${this._successCount}.\n`);
        }
    }
    /**
     * Get a flattened list of all samples relevant to this dataset converter
     * @returns All relevant samples
     */
    getSamples() {
        return this._samples;
    }
    /**
     * Get the converted annotations for a sample
     * @param sample Sample to get annotations for
     * @returns Sample annotations (bounding boxes or label) if they can be found
     */
    async getAnnotationsForSample(sample) {
        if (!this._categoryAnnotations || !this._directoryAnnotations || !this._samples) {
            throw new Error('Attempting to get sample annotations before dataset has been converted');
        }
        // Get the sample annotation
        let annotation = this.getAnnotation(sample.filename, sample.category, sample.directory);
        if (annotation && this._datasetFormat.type === 'object-detection'
            && this._datasetFormat.additionalOpts) {
            const samplePath = path_1.default.join(sample.directory, sample.filename);
            annotation = this.transformAnnotations(annotation, samplePath);
        }
        return {
            label: annotation && this._datasetFormat.type === 'single-label' ?
                {
                    type: 'label',
                    label: annotation,
                } : { type: 'infer' },
            boundingBoxes: annotation && this._datasetFormat.type === 'object-detection' ?
                annotation : undefined,
        };
    }
    /**
     * Process a single file: if this file is an annotations file, parse and store it. Otherwise, add it to the set
     * of files to be uploaded.
     * @param filename File name
     * @param path File path
     */
    parseFile(filename, path) {
        var _a;
        if (!this._categoryAnnotations || !this._directoryAnnotations || !this._samples) {
            throw new Error('No annotations yet');
        }
        const filePath = path_1.default.join(path, filename);
        try {
            if (((_a = this._datasetFormat.additionalOpts) === null || _a === void 0 ? void 0 : _a.needsLabelmap) && this._datasetFormat.labelsFile
                && this._datasetFormat.labelsFile.find(format => format.fileFilter(filename))) {
                // Store a label map file
                this._labelMapFiles.push({
                    path: filePath,
                    name: filename,
                });
            }
            else if (this._datasetFormat.fileFilter(filename)) {
                // This file is potentially a label file
                let labelFileRes;
                const fileContents = fs_1.default.readFileSync(filePath, { encoding: 'utf-8' });
                if (this._datasetFormat.formatStyle === 'json') {
                    const jsonParseRes = parse_label_json_1.parseJsonLabelsFile(fileContents, this._datasetFormat.key);
                    if (jsonParseRes.success) {
                        labelFileRes = {
                            type: jsonParseRes.type,
                            labels: jsonParseRes.labels,
                        };
                    }
                    else {
                        throw new Error(jsonParseRes.reason);
                    }
                }
                else if (this._datasetFormat.formatStyle === 'csv') {
                    // Parse the text (as CSV) into rows and columns
                    const csvParsed = papaparse_1.default.parse(fileContents, {
                        delimitersToGuess: [',', '\t', ';'],
                    });
                    const data = csvParsed.data;
                    // Check if the columns match any known schema, and parse the rows
                    const csvParseRes = parse_label_csv_1.parseCsvLabelsFile(data, this._datasetFormat.key);
                    if (csvParseRes.success) {
                        labelFileRes = {
                            type: csvParseRes.type,
                            labels: csvParseRes.labels,
                        };
                    }
                    else {
                        throw new Error(csvParseRes.reason);
                    }
                }
                else if (this._datasetFormat.formatStyle === 'txt') {
                    const txtParseRes = parse_label_json_1.parseTxtLabelsFile(fileContents, filename, this._datasetFormat.key);
                    if (txtParseRes.success) {
                        labelFileRes = {
                            type: txtParseRes.type,
                            labels: txtParseRes.labels,
                        };
                    }
                    else {
                        throw new Error(txtParseRes.reason);
                    }
                }
                else if (this._datasetFormat.formatStyle === 'bounding_boxes.labels') {
                    // Original bounding boxes format
                    const bboxFile = JSON.parse(fileContents);
                    labelFileRes = {
                        type: 'object-detection',
                        labels: bounding_box_file_types_1.validateBoundingBoxLabels(bboxFile).boundingBoxes,
                    };
                }
                else if (this._datasetFormat.formatStyle === 'xml') {
                    // Parse the text into an XMLDocument
                    const xmlParser = new xmldom_1.DOMParser();
                    const xmlParsed = xmlParser.parseFromString(fileContents, 'application/xml');
                    // Check if the XMLDocument matches any known schema
                    const xmlParseRes = parse_label_xml_1.parseXmlLabelsFile(xmlParsed, this._datasetFormat.key);
                    if (xmlParseRes.success) {
                        labelFileRes = {
                            type: xmlParseRes.type,
                            labels: xmlParseRes.labels,
                        };
                    }
                    else {
                        throw new Error(xmlParseRes.reason);
                    }
                }
                else {
                    throw new Error('File has unsupported file type');
                }
                // If we can, we'll associate these annotations with a category.
                // We try to derive this first from the directory path, then the file name.
                const directoryCategory = label_files_shared_1.findCategoryMatch(path);
                const fileCategory = label_files_shared_1.findCategoryMatch(filename);
                this.storeAnnotations(directoryCategory, fileCategory, path, labelFileRes);
                this._successCount++;
                if (!this._opts.silent) {
                    console.log(`Parsed annotations file '${filename}' successfully`);
                }
            }
            else if (this._opts.validExtensions.indexOf(path_1.default.extname(filename.toLowerCase())) !== -1) {
                // A non-label file; work out its category and store
                const directoryCategory = label_files_shared_1.findCategoryMatch(path);
                this._samples.push({
                    filename: filename,
                    directory: path,
                    category: directoryCategory
                });
            }
            else {
                // We cannot handle this file; ignore it
                throw new Error(`File is not a valid label file (expected ${this._datasetFormat.formatStyle}) ` +
                    `or valid sample file (unsupported extension)`);
            }
        }
        catch (ex) {
            if (!this._opts.silent) {
                console.log(`Skipping file '${filePath}': ${ex}`);
            }
        }
    }
    /**
     * Attempt to automatically find and parse a label map file.
     * @returns Label map
     */
    async getLabelmap() {
        var _a;
        if (!((_a = this._datasetFormat.additionalOpts) === null || _a === void 0 ? void 0 : _a.needsLabelmap) || !this._datasetFormat.labelsFile) {
            return;
        }
        let labelMap;
        // First try to automatically find and parse a labels file
        for (const labelsFileFormat of this._datasetFormat.labelsFile) {
            const labelFile = this._labelMapFiles.find(f => labelsFileFormat.fileFilter(f.name));
            if (!labelFile) {
                continue;
            }
            try {
                const fileContents = fs_1.default.readFileSync(labelFile.path, { encoding: 'utf8' });
                if (labelsFileFormat.fileType === 'txt') {
                    labelMap = labelsFileFormat.conversionFunction(fileContents);
                }
                else if (labelsFileFormat.fileType === 'csv') {
                    const csvParsed = papaparse_1.default.parse(fileContents, {
                        delimitersToGuess: [',', '\t', ';'],
                    });
                    const data = csvParsed.data;
                    labelMap = labelsFileFormat.conversionFunction(data);
                }
                else if (labelsFileFormat.fileType === 'yaml') {
                    const yamlParsed = js_yaml_1.default.load(fileContents);
                    labelMap = labelsFileFormat.conversionFunction(yamlParsed);
                }
                if (labelMap) {
                    console.log(`Successfully parsed label map file '${labelFile.name}'`);
                    return labelMap;
                }
            }
            catch (ex) {
                console.log(`Failed to automatically parse label map file '${labelFile.name}'`, ex);
            }
        }
        console.log('\nA label map file could not be detected.');
        console.log('This format requires a label map file, which maps keys to the label they represent.');
        console.log('You can fix these labels later by clicking "Edit labels" on the data acquisition page.\n');
        const continueRes = await inquirer_1.default.prompt({
            type: 'confirm',
            name: 'continue',
            message: 'Continue?'
        });
        if (!continueRes.continue) {
            process.exit(0);
        }
    }
    /**
     * Transform bounding boxes (e.g. when normalized)
     * @param annotations Annotations to transform
     * @param filepath File these annotations are specific to
     * @returns Transformed annotations, if relevant for this dataset format
     */
    transformAnnotations(annotations, filepath) {
        const opts = this._datasetFormat.additionalOpts;
        if (!opts)
            return annotations;
        let width;
        let height;
        if (opts.normalizedBoundingBoxes) {
            const dimensions = image_size_1.default(filepath);
            width = dimensions.width;
            height = dimensions.height;
        }
        return this.transformBoundingBoxes(annotations, width, height);
    }
}
exports.DatasetConverterHelperCli = DatasetConverterHelperCli;
/**
 * Get all files in a folder (including sub-folders)
 * @param folder Path to folder to read
 */
function getAllFilesInFolder(folder) {
    let allFiles = [];
    fs_helpers_1.FSHelpers.readAllFiles(folder, (file) => allFiles.push(file));
    return allFiles;
}
exports.getAllFilesInFolder = getAllFilesInFolder;
/**
 * Given a set of files, try to find a match against a known dataset format.
 * @param files Set of files to analyze
 * @param formatsToTry List of formats to try; if not given, will try all formats
 * @returns Dataset format if a match can be found
 */
function deriveDatasetFormat(files, formatsToTry) {
    // Only try to find dataset format if we've uploaded any images
    const someImageFiles = files.some(f => {
        const filename = f.filename.toLowerCase();
        return filename.endsWith('.jpg') || filename.endsWith('.jpeg') || filename.endsWith('.png');
    });
    if (!someImageFiles) {
        return undefined;
    }
    // Examine each image label file, check if it matches a known type
    const imageLabelFiles = files.filter(f => f.filename.toLowerCase().endsWith('.csv')
        || f.filename.toLowerCase().endsWith('.json')
        || f.filename.toLowerCase().endsWith('.xml')
        || f.filename.toLocaleLowerCase() === 'bounding_boxes.labels');
    // If we filter to a list of formats to try, use only those formats
    let formatsToTryJson;
    let formatsToTryCsv;
    let formatsToTryXml;
    let tryBboxes = true;
    if (formatsToTry) {
        formatsToTryJson = label_file_types_1.supportedJsonLabelFormats.filter(format => formatsToTry.includes(format.info.key));
        formatsToTryCsv = label_file_types_1.supportedCsvLabelFormats.filter(format => formatsToTry.includes(format.info.key));
        formatsToTryXml = label_file_types_1.supportedXmlLabelFormats.filter(format => formatsToTry.includes(format.info.key));
        tryBboxes = formatsToTry.includes('ei-bounding-boxes-format');
    }
    for (const labelFile of imageLabelFiles) {
        const filePath = path_1.default.join(labelFile.path, labelFile.filename);
        const labelFileText = fs_1.default.readFileSync(filePath, { encoding: 'utf-8' });
        const labelFileName = labelFile.filename.toLowerCase();
        if (labelFileName.endsWith('.json')) {
            // Look for any JSON format match
            const format = parse_label_json_1.checkJsonMatchesAnyFormat(labelFileText, formatsToTryJson);
            if (format) {
                return format.key;
            }
        }
        else if (labelFileName.endsWith('.csv')) {
            // Parse the text (as CSV) into rows and columns
            const csvParsed = papaparse_1.default.parse(labelFileText, {
                delimitersToGuess: [',', '\t', ';'],
            });
            const data = csvParsed.data;
            // Look for any CSV format match
            const format = parse_label_csv_1.checkCsvMatchesAnyFormat(data, formatsToTryCsv);
            if (format) {
                return format.key;
            }
        }
        else if (labelFileName.endsWith('.xml')) {
            // Parse the text into an XMLDocument
            const xmlParser = new xmldom_1.DOMParser();
            const xmlParsed = xmlParser.parseFromString(labelFileText, 'application/xml');
            // Look for any XML format match
            const format = parse_label_xml_1.checkXmlMatchesAnyFormat(xmlParsed, formatsToTryXml);
            if (format) {
                return format.key;
            }
        }
        else if (labelFileName === 'bounding_boxes.labels' && tryBboxes) {
            // Original bounding boxes format
            return 'ei-bounding-boxes-format';
        }
    }
    // No match for any label file
    return undefined;
}
exports.deriveDatasetFormat = deriveDatasetFormat;
/**
 * Check a dataset matches a given format
 * @param files Set of files to analyze
 * @param formatName Format to check
 * @returns True if the dataset matches the format, else false
 */
function checkDatasetMatchesFormat(files, formatName) {
    const format = deriveDatasetFormat(files, [formatName]);
    if (format && format === formatName) {
        return true;
    }
    else {
        return false;
    }
}
exports.checkDatasetMatchesFormat = checkDatasetMatchesFormat;
//# sourceMappingURL=dataset-converter-cli.js.map