#!/usr/bin/env node
"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const serial_connector_1 = require("../serial-connector");
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const eta_serial_protocol_1 = __importDefault(require("./eta-serial-protocol"));
const find_serial_1 = require("../find-serial");
const crc_32_1 = __importDefault(require("crc-32"));
const commander_1 = __importDefault(require("commander"));
const cli_progress_1 = __importDefault(require("cli-progress"));
const check_new_version_1 = __importDefault(require("../check-new-version"));
const config_1 = require("../config");
const SERIAL_PREFIX = '\x1b[33m[ETA]\x1b[0m';
const APP_BAUD_RATE = 115200;
const packageVersion = JSON.parse(fs_1.default.readFileSync(path_1.default.join(__dirname, '..', '..', '..', 'package.json'), 'utf-8')).version;
commander_1.default
    .description('Eta Compute ECM3532 AI Sensor flash tool')
    .version(packageVersion)
    .option('-f --firmware-path <file>', 'Firmware path (required)')
    .option('--baud-rate <n>', 'Bootloader baud rate (default: 460800)')
    .option('--serial-write-pacing', 'Enable write pacing (default: off)')
    .option('--skip-reset', 'Skip the reset procedure (in case the device is already in bootloader mode)')
    .option('--verbose', 'Enable debug logs')
    .allowUnknownOption(true)
    .parse(process.argv);
const firmwarePathArgv = commander_1.default.firmwarePath;
const baudRateArgv = commander_1.default.baudRate;
const serialWritePacing = !!commander_1.default.serialWritePacing;
const debug = !!commander_1.default.verbose;
const skipReset = !!commander_1.default.skipReset;
let configFactory = new config_1.Config();
// tslint:disable-next-line:no-floating-promises
(async () => {
    try {
        if (!firmwarePathArgv) {
            console.log('Missing --firmware-path argument');
            process.exit(1);
        }
        if (!fs_1.default.existsSync(firmwarePathArgv)) {
            console.log(firmwarePathArgv + ' does not exist (via --firmware-path)');
            process.exit(1);
        }
        try {
            await check_new_version_1.default(configFactory);
        }
        catch (ex) {
            /* noop */
        }
        let deviceId = await find_serial_1.findSerial(undefined);
        await connectToSerial(deviceId);
    }
    catch (ex) {
        console.error('Failed to set up serial daemon', ex);
    }
})();
function sleep(ms) {
    return new Promise((res) => setTimeout(res, ms));
}
async function connectToSerial(deviceId) {
    const bootloaderBaudRate = baudRateArgv ? Number(baudRateArgv) : 460800;
    if (isNaN(bootloaderBaudRate)) {
        throw new Error('Baud rate is invalid (' + baudRateArgv + ', via --baud-rate)');
    }
    // if this is set we have a connection to the server
    let ws;
    console.log(SERIAL_PREFIX, 'Connecting to ' + deviceId + '...');
    const serial = new serial_connector_1.SerialConnector(deviceId, APP_BAUD_RATE);
    const serialProtocol = new eta_serial_protocol_1.default(serial, serialWritePacing, debug);
    serial.on('error', (err) => {
        console.log(SERIAL_PREFIX, 'Serial error - retrying in 5 seconds', err.message || err.toString());
        setTimeout(serial_connect, 5000);
    });
    serial.on('close', () => {
        console.log(SERIAL_PREFIX, 'Serial closed - retrying in 5 seconds');
        if (ws) {
            ws.terminate();
        }
        setTimeout(serial_connect, 5000);
    });
    // serial.on('data', data => {
    //     console.log(SERIAL_PREFIX, 'serial data', data, data.toString('ascii'));
    //     // client.write(data);
    // });
    async function connectLogic() {
        if (!serial.isConnected())
            return setTimeout(serial_connect, 5000);
        if (!await serial.hasSerial()) {
            console.log('');
            console.log('Lost connection to the device. Are you connected to the right serial port?');
            process.exit(1);
        }
        const progressBar = new cli_progress_1.default.SingleBar({}, cli_progress_1.default.Presets.shades_classic);
        try {
            if (!skipReset) {
                let waitForBootloaderTimeout = 60000;
                console.log(SERIAL_PREFIX, 'Connected, detecting firmware...');
                try {
                    await serialProtocol.onConnected();
                    await serialProtocol.rebootIntoBootloader();
                    waitForBootloaderTimeout = 10000;
                }
                catch (ex2) {
                    console.log(SERIAL_PREFIX, 'Press the **RESET** button on your AI Sensor');
                }
                await serialProtocol.setBaudRate(bootloaderBaudRate);
                await serialProtocol.waitForBootloader(waitForBootloaderTimeout);
            }
            else {
                console.log(SERIAL_PREFIX, '--skip-reset active, assuming bootloader is loaded');
                await serialProtocol.setBaudRate(bootloaderBaudRate);
            }
            let version = await serialProtocol.getVersion();
            console.log(SERIAL_PREFIX, 'Restarted into bootloader. Device is running bootloader version ' +
                version.major + '.' + version.minor + '.' +
                version.patch);
            // we support devices with version 1.0.x and lower
            if (isNaN(version.major) || isNaN(version.minor) || version.major > 1 || version.minor > 0) {
                console.error(SERIAL_PREFIX, 'Unsupported bootloader version running on this device. Supported version is 1.0.x and lower, ' +
                    'but found ' + version.major + '.' + version.minor + '.' +
                    version.patch + '.');
                console.error(SERIAL_PREFIX, 'Update the Eta Compute Flash tool ' +
                    'to continue via `npm install -g edge-impulse-cli@latest`.');
                process.exit(1);
            }
            let firmware = await fs_1.default.promises.readFile(firmwarePathArgv || '');
            let fullHash = crc_32_1.default.buf(firmware);
            let blockLength = await serialProtocol.sendAppInfo('New firmware', 'ECM3532 M3', fullHash, firmware.length);
            if (version.patch < 2) {
                blockLength = 255;
            }
            let blocks = [];
            for (let ix = 0; ix < firmware.length; ix += blockLength) {
                blocks.push(firmware.slice(ix, ix + blockLength));
            }
            console.log(SERIAL_PREFIX, 'Firmware is', firmware.length, 'bytes (' + blocks.length + ' blocks)');
            await sleep(1000);
            if (!debug) {
                progressBar.start(blocks.length, 0);
            }
            let blockIx = 0;
            for (let b of blocks) {
                blockIx++;
                let currBlock = (blockIx).toString().padStart(blocks.length.toString().length, ' ');
                let currBlockHash = crc_32_1.default.buf(b);
                if (debug) {
                    console.log(SERIAL_PREFIX, '[' + currBlock + '/' + blocks.length + '] ' +
                        'Sending block (hash ' + currBlockHash + ')...');
                }
                await serialProtocol.sendBlock(blockIx, b, currBlockHash);
                if (!debug) {
                    progressBar.update(blockIx);
                }
            }
            if (!debug) {
                progressBar.stop();
            }
            console.log(SERIAL_PREFIX, 'Finalizing update...');
            await serialProtocol.sendEou();
            console.log(SERIAL_PREFIX, 'Update complete, rebooting into application');
            await serialProtocol.rebootIntoApplication();
            console.log(SERIAL_PREFIX, 'Update completed');
            process.exit(0);
        }
        catch (ex) {
            console.error(SERIAL_PREFIX, 'Error during update.', ex);
            console.error(SERIAL_PREFIX, 'Is this device running the Eta Compute bootloader and did you press the **RESET** button?');
            process.exit(1);
        }
        finally {
            progressBar.stop();
        }
    }
    serial.on('connected', connectLogic);
    async function serial_connect() {
        try {
            await serial.connect();
        }
        catch (ex2) {
            let ex = ex2;
            console.error(SERIAL_PREFIX, 'Failed to connect to', deviceId, 'retrying in 5 seconds', ex.message || ex);
            if (ex.message && ex.message.indexOf('Permission denied')) {
                console.error(SERIAL_PREFIX, 'You might need `sudo` or set up the right udev rules');
            }
            setTimeout(serial_connect, 5000);
        }
    }
    // tslint:disable-next-line:no-floating-promises
    serial_connect();
}
//# sourceMappingURL=flashtool.js.map