"use strict";
// tslint:disable: no-unsafe-any
Object.defineProperty(exports, "__esModule", { value: true });
exports.SerialConnector = void 0;
// tslint:disable-next-line
const serialPort = require('serialport');
const events_1 = require("events");
// Don't open same port twice
let serialPorts = {};
class SerialConnector extends events_1.EventEmitter {
    constructor(path, baudrate, echoSerial = false) {
        super();
        this._echoSerial = echoSerial;
        this._path = path;
        this._baudrate = baudrate;
        this._dataHandler = (d) => {
            if (this._echoSerial) {
                const CON_PREFIX = '\x1b[36m[Rx ]\x1b[0m';
                console.log(CON_PREFIX, d.toString('ascii'));
            }
            this.emit('data', d);
        };
        this._connected = false;
    }
    static async list() {
        return (await serialPort.list());
    }
    getPath() {
        return this._path;
    }
    isConnected() {
        return this._connected;
    }
    async connect() {
        let alreadyConnected = false;
        if (serialPorts[this._path]) {
            alreadyConnected = true;
            this._serial = serialPorts[this._path];
        }
        else {
            this._serial = new serialPort(this._path, { baudRate: this._baudrate });
            serialPorts[this._path] = this._serial;
            this._serial.on('close', () => {
                this._serial = null;
                delete serialPorts[this._path];
                this._connected = false;
                this.emit('close');
            });
        }
        this._serial.on('data', this._dataHandler);
        if (alreadyConnected) {
            return;
        }
        // otherwise wait for either error or open event
        return new Promise((resolve, reject) => {
            this._serial.once('error', (ex) => {
                this._serial = null;
                delete serialPorts[this._path];
                reject(ex);
            });
            this._serial.once('open', () => {
                this._connected = true;
                this.emit('connected');
                this._serial.on('error', (ex) => this.emit('error', ex));
                resolve();
            });
        });
    }
    async write(buffer) {
        return new Promise((res, rej) => {
            if (!this._serial)
                return rej('Serial is null');
            if (this._echoSerial) {
                const CON_PREFIX = '\x1b[35m[Tx ]\x1b[0m';
                console.log(CON_PREFIX, buffer.toString('ascii'));
            }
            this._serial.write(buffer, (err) => {
                if (err)
                    return rej(err);
                res();
            });
        });
    }
    async setBaudRate(baudRate) {
        await this._serial.update({
            baudRate: baudRate
        });
        this._baudrate = baudRate;
    }
    getBaudRate() {
        return this._baudrate;
    }
    async disconnect() {
        this._serial.off('data', this._dataHandler);
        return true;
    }
    async getMACAddress() {
        let list = await SerialConnector.list();
        let l = list.find(j => j.path === this._path);
        return l ? l.serialNumber : null;
    }
    async hasSerial() {
        return !!this._serial;
    }
    canSwitchBaudRate() {
        return true;
    }
}
exports.SerialConnector = SerialConnector;
//# sourceMappingURL=serial-connector.js.map