"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.allFormatsWithSubformats = exports.listAllAnnotationFormats = exports.getMetadataForFormat = exports.supportedXmlLabelFormats = exports.supportedTxtLabelFormats = exports.supportedCsvLabelFormats = exports.supportedJsonLabelFormats = void 0;
const label_json_schema_1 = require("./label-json-schema");
// Supported JSON formats
exports.supportedJsonLabelFormats = [
    // COCO
    {
        info: {
            key: 'coco',
            name: 'COCO JSON',
            parentFormat: 'coco-json',
        },
        fn: {
            fileType: 'json',
            type: 'object-detection',
            conversionFunction: (json) => {
                const jsonParsed = json;
                // Map image IDs to filenames
                let filenameMap = new Map();
                for (const image of jsonParsed.images) {
                    filenameMap.set(image.id, image.file_name);
                }
                // Map category IDs to labels
                let categoryMap = new Map();
                for (const category of jsonParsed.categories) {
                    categoryMap.set(category.id, category.name);
                }
                // Now extract all bounding boxes
                let bboxes = {};
                for (const annotation of jsonParsed.annotations) {
                    const filename = filenameMap.get(annotation.image_id);
                    if (!filename) {
                        throw new Error(`Missing file info for image ID ${annotation.image_id}`);
                    }
                    const label = categoryMap.get(annotation.category_id);
                    if (!label) {
                        throw new Error(`Missing category info for category ID ${annotation.category_id}`);
                    }
                    // May or may not be nested
                    let allBoxes = [];
                    if (!Array.isArray(annotation.bbox[0])) {
                        const box = annotation.bbox;
                        allBoxes.push(box);
                    }
                    else {
                        allBoxes = annotation.bbox;
                    }
                    const parsedBoxes = allBoxes.map(box => {
                        if (box.length !== 4) {
                            throw new Error(`Expected bounding box to have length 4 but got ${box.length}`);
                        }
                        return {
                            label: label,
                            x: box[0],
                            y: box[1],
                            width: box[2],
                            height: box[3]
                        };
                    });
                    if (bboxes[filename]) {
                        bboxes[filename] = [...bboxes[filename], ...parsedBoxes];
                    }
                    else {
                        bboxes[filename] = parsedBoxes;
                    }
                }
                return {
                    type: 'object-detection',
                    bboxes: bboxes,
                };
            },
            schema: label_json_schema_1.cocoJsonSchema,
        },
    },
    // Remo (single label)
    {
        info: {
            key: 'remo-json-single-label',
            name: 'Remo JSON (single label)',
            parentFormat: 'remo-json',
        },
        fn: {
            fileType: 'json',
            type: 'single-label',
            conversionFunction: (json) => {
                const jsonParsed = json;
                let labels = {};
                jsonParsed.forEach(row => {
                    labels[row.file_name] = row.classes[0];
                });
                return {
                    type: 'single-label',
                    labels: labels,
                };
            },
            schema: label_json_schema_1.remoSingleLabelSchema,
        }
    },
    // Remo (object detection)
    {
        info: {
            key: 'remo-json-object-detection',
            name: 'Remo JSON (object detection)',
            parentFormat: 'remo-json',
        },
        fn: {
            fileType: 'json',
            type: 'object-detection',
            conversionFunction: (json) => {
                const jsonParsed = json;
                let bboxes = {};
                for (const file of jsonParsed) {
                    bboxes[file.file_name] = [];
                    for (const annotation of file.annotations) {
                        const box = annotation.bbox;
                        for (const classname of annotation.classes) {
                            bboxes[file.file_name].push({
                                label: classname,
                                x: box.xmin,
                                y: box.ymin,
                                width: box.xmax - box.xmin,
                                height: box.ymax - box.ymin
                            });
                        }
                    }
                }
                return {
                    type: 'object-detection',
                    bboxes: bboxes
                };
            },
            schema: label_json_schema_1.remoObjectDetectionSchema,
        }
    },
];
const plainCsvObjDetSchema = ['file_name', 'classes', 'xmin', 'ymin', 'xmax', 'ymax'];
const plainCsvSchema = ['file_name', 'class_name'];
const openImagesCsvSchema = ['ImageID', 'LabelName', 'XMin', 'XMax', 'YMin', 'YMax'];
// Supported formats that come in a CSV file
exports.supportedCsvLabelFormats = [
    // Plain CSV (object detection)
    {
        info: {
            key: 'plain-csv-object-detection',
            name: 'Plain CSV (object detection)',
            parentFormat: 'plain-csv',
        },
        fn: {
            fileType: 'csv',
            type: 'object-detection',
            schema: Array.from(plainCsvObjDetSchema),
            conversionFunction: (csv) => {
                let bboxes = {};
                for (const row of csv) {
                    const filename = row.file_name;
                    const classes = row.classes.split(';').map(s => s.trim());
                    const x = Number(row.xmin);
                    const y = Number(row.ymin);
                    const xMax = Number(row.xmax);
                    const yMax = Number(row.ymax);
                    if (isNaN(x) || isNaN(y) || isNaN(xMax) || isNaN(yMax)) {
                        continue;
                    }
                    const width = xMax - x;
                    const height = yMax - y;
                    if (width < 0) {
                        continue;
                    }
                    if (height < 0) {
                        continue;
                    }
                    for (const className of classes) {
                        if (!bboxes[filename]) {
                            bboxes[filename] = [];
                        }
                        bboxes[filename].push({
                            label: className,
                            x: x,
                            y: y,
                            width: width,
                            height: height,
                        });
                    }
                }
                return {
                    type: 'object-detection',
                    bboxes: bboxes,
                };
            }
        },
    },
    // Plain CSV (single label)
    {
        info: {
            key: 'plain-csv',
            name: 'Plain CSV (single label)',
            parentFormat: 'plain-csv',
        },
        fn: {
            fileType: 'csv',
            type: 'single-label',
            schema: Array.from(plainCsvSchema),
            conversionFunction: (csv) => {
                let labels = {};
                for (const row of csv) {
                    labels[row.file_name] = row.class_name;
                }
                return {
                    type: 'single-label',
                    labels: labels,
                };
            }
        }
    },
    // Open images CSV
    {
        info: {
            key: 'open-images-csv',
            name: 'Open Images CSV',
            parentFormat: 'open-images-csv',
        },
        fn: {
            fileType: 'csv',
            type: 'object-detection',
            schema: Array.from(openImagesCsvSchema),
            conversionFunction: (csv) => {
                let bboxes = {};
                for (const row of csv) {
                    const filename = row.ImageID;
                    const className = row.LabelName;
                    const x = Number(row.XMin);
                    const y = Number(row.YMin);
                    const xMax = Number(row.XMax);
                    const yMax = Number(row.YMax);
                    if (isNaN(x) || isNaN(y) || isNaN(xMax) || isNaN(yMax)) {
                        continue;
                    }
                    const width = xMax - x;
                    const height = yMax - y;
                    if (width < 0) {
                        continue;
                    }
                    if (height < 0) {
                        continue;
                    }
                    if (!bboxes[filename]) {
                        bboxes[filename] = [];
                    }
                    bboxes[filename].push({
                        label: className,
                        x: x,
                        y: y,
                        width: width,
                        height: height,
                    });
                }
                return {
                    type: 'object-detection',
                    bboxes: bboxes,
                };
            },
            additionalOpts: {
                needsLabelmap: true,
                normalizedBoundingBoxes: true,
                centeredBoundingBoxes: false,
                lookupNoExtension: true,
            },
            labelsFile: [{
                    fileFilter: (name) => {
                        return name.endsWith('class-descriptions.csv');
                    },
                    fileType: 'csv',
                    conversionFunction: (csv) => {
                        let labels = {};
                        for (const row of csv) {
                            if (row.length < 2)
                                continue;
                            labels[row[0]] = row[1];
                        }
                        return labels;
                    },
                    fileDescription: 'class-descriptions.csv',
                    isNumeric: false
                }]
        }
    }
];
// Supported formats that come in a txt file
exports.supportedTxtLabelFormats = [
    {
        info: {
            key: 'yolo-txt',
            name: 'YOLO TXT',
            parentFormat: 'yolo-txt',
        },
        fn: {
            fileType: 'txt',
            type: 'object-detection',
            conversionFunction: (txt) => {
                const lines = txt.split('\n');
                let boxes = [];
                for (let i = 0; i < lines.length; i++) {
                    const line = lines[i];
                    const lineParts = line.split(' ');
                    if (lineParts.length !== 5) {
                        continue;
                    }
                    const label = lineParts[0];
                    const x = Number(lineParts[1]);
                    const y = Number(lineParts[2]);
                    const width = Number(lineParts[3]);
                    const height = Number(lineParts[4]);
                    if (isNaN(x) || isNaN(y) || isNaN(width) || isNaN(height)) {
                        throw new Error(`Got NaN value on line ${i}`);
                    }
                    boxes.push({
                        x: x,
                        y: y,
                        width: width,
                        height: height,
                        label: label
                    });
                }
                if (boxes.length === 0) {
                    throw new Error('No valid boxes found');
                }
                return boxes;
            },
            additionalOpts: {
                needsLabelmap: true,
                normalizedBoundingBoxes: true,
                centeredBoundingBoxes: true,
                lookupNoExtension: true,
            },
            labelsFile: [
                {
                    fileFilter: (name) => {
                        return name.toLowerCase().endsWith('.names')
                            || name.toLowerCase() === 'classes.txt';
                    },
                    fileType: 'txt',
                    conversionFunction: (txt) => {
                        let labelMap = {};
                        const rows = txt.split('\n');
                        for (let i = 0; i < rows.length; i++) {
                            labelMap[i] = rows[i];
                        }
                        return labelMap;
                    },
                    fileDescription: 'labels.names',
                    isNumeric: true
                },
                {
                    fileFilter: (name) => {
                        return name.toLowerCase() === 'data.yaml';
                    },
                    fileType: 'yaml',
                    conversionFunction: (yamlContents) => {
                        let labelMap = {};
                        // tslint:disable-next-line: no-unsafe-any
                        const yamlNames = yamlContents.names;
                        if (typeof yamlNames === 'undefined') {
                            throw new Error('File contains no "names" entry');
                        }
                        if (!Array.isArray(yamlNames)) {
                            throw new Error('"names" is not an array');
                        }
                        let nameIx = 0;
                        for (const name of yamlNames) {
                            if (typeof name === 'string') {
                                labelMap[nameIx++] = name;
                            }
                        }
                        return labelMap;
                    },
                    fileDescription: 'labels.names',
                    isNumeric: true
                },
            ]
        }
    }
];
exports.supportedXmlLabelFormats = [
    {
        info: {
            key: 'pascal-voc-xml-object-detection',
            name: 'Pascal VOC XML (object detection)',
            parentFormat: 'pascal-voc-xml',
        },
        fn: {
            fileType: 'xml',
            conversionFunction: (json) => {
                const jsonParsed = json;
                let bboxes = {};
                const boxesToProcess = Array.isArray(jsonParsed.annotation.object)
                    ? jsonParsed.annotation.object : [jsonParsed.annotation.object];
                let boxes = [];
                // Process each box
                for (const box of boxesToProcess) {
                    const x = Number(box.bndbox.xmin);
                    if (isNaN(x) || x < 0) {
                        throw new Error(`Got invalid value for xmin: ${box.bndbox.xmin}`);
                    }
                    const y = Number(box.bndbox.ymin);
                    if (isNaN(y) || y < 0) {
                        throw new Error(`Got invalid value for ymin: ${box.bndbox.ymin}`);
                    }
                    const xmax = Number(box.bndbox.xmax);
                    if (isNaN(xmax) || xmax < 0) {
                        throw new Error(`Got invalid value for xmax: ${box.bndbox.xmax}`);
                    }
                    if (xmax < x) {
                        throw new Error('Got bounding box with negative width');
                    }
                    const ymax = Number(box.bndbox.ymax);
                    if (isNaN(ymax) || ymax < 0) {
                        throw new Error(`Got invalid value for ymax: ${box.bndbox.ymax}`);
                    }
                    if (ymax < y) {
                        throw new Error('Got bounding box with negative height');
                    }
                    boxes.push({
                        x: x,
                        y: y,
                        width: xmax - x,
                        height: ymax - y,
                        label: box.name,
                    });
                }
                // Store boxes against this filename
                bboxes[jsonParsed.annotation.filename] = boxes;
                return {
                    type: 'object-detection',
                    bboxes: bboxes,
                };
            },
            schema: label_json_schema_1.pascalVocSchema,
            type: 'object-detection'
        }
    }
];
const allSupportedFormats = [
    ...exports.supportedJsonLabelFormats,
    ...exports.supportedCsvLabelFormats,
    ...exports.supportedTxtLabelFormats,
    ...exports.supportedXmlLabelFormats,
].sort((a, b) => a.info.name.localeCompare(b.info.name));
const allFormatsMap = new Map();
allSupportedFormats.forEach(format => allFormatsMap.set(format.info.key, format));
/**
 * Get info about a dataset format type
 * @param format Dataset format; if we don't recognize it, return undefined
 * @returns Dataset format metadata, or undefined
 */
function getMetadataForFormat(format) {
    if (!format)
        return undefined;
    if (format === 'ei-bounding-boxes-format') {
        return {
            key: 'ei-bounding-boxes-format',
            type: 'object-detection',
            fileFilter: (name) => {
                return name.toLowerCase() === 'bounding_boxes.labels';
            },
            formatStyle: 'bounding_boxes.labels',
            additionalOpts: undefined,
        };
    }
    const formatMatch = allFormatsMap.get(format);
    if (formatMatch) {
        return {
            key: formatMatch.info.key,
            type: formatMatch.fn.type,
            fileFilter: (name) => {
                return name.toLowerCase().endsWith(`.${formatMatch.fn.fileType}`);
            },
            formatStyle: formatMatch.fn.fileType,
            additionalOpts: formatMatch.fn.additionalOpts,
            labelsFile: formatMatch.fn.labelsFile
        };
    }
    // Unrecognized format
    return undefined;
}
exports.getMetadataForFormat = getMetadataForFormat;
/**
 * Get info for all dataset annotation formats we can convert from
 * @returns Names and keys of all formats we support
 */
function listAllAnnotationFormats(filter) {
    const eiBoundingBoxesInfo = {
        name: 'Edge Impulse object detection dataset',
        key: 'ei-bounding-boxes-format',
        type: 'object-detection',
    };
    return [
        eiBoundingBoxesInfo,
        ...allSupportedFormats.map(format => {
            return {
                name: format.info.name,
                key: format.info.key,
                type: format.fn.type,
            };
        })
    ].filter(format => !filter || format.type === filter);
}
exports.listAllAnnotationFormats = listAllAnnotationFormats;
exports.allFormatsWithSubformats = {
    'edge-impulse-format': {
        name: 'Edge Impulse object detection dataset',
        subformats: [
            {
                type: 'object-detection',
                key: 'ei-bounding-boxes-format'
            }
        ]
    },
    'coco-json': {
        name: 'COCO JSON',
        subformats: [
            {
                type: 'object-detection',
                key: 'coco'
            }
        ]
    },
    'open-images-csv': {
        name: 'Open Images CSV',
        subformats: [
            {
                type: 'object-detection',
                key: 'open-images-csv'
            }
        ]
    },
    'pascal-voc-xml': {
        name: 'Pascal VOC XML',
        subformats: [
            {
                type: 'object-detection',
                key: 'pascal-voc-xml-object-detection'
            }
        ]
    },
    'plain-csv': {
        name: 'Plain CSV',
        subformats: [
            {
                type: 'object-detection',
                key: 'plain-csv-object-detection'
            },
            {
                type: 'single-label',
                key: 'plain-csv'
            }
        ]
    },
    'remo-json': {
        name: 'Remo JSON',
        subformats: [
            {
                type: 'object-detection',
                key: 'remo-json-object-detection'
            },
            {
                type: 'single-label',
                key: 'remo-json-single-label'
            }
        ]
    },
    'yolo-txt': {
        name: 'YOLO TXT',
        subformats: [
            {
                type: 'object-detection',
                key: 'yolo-txt'
            }
        ]
    }
};
//# sourceMappingURL=label-file-types.js.map