"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseTxtLabelsFile = exports.parseJsonLabelsFile = exports.checkJsonMatchesAnyFormat = exports.validateJsonSchema = exports.checkJsonMatchesFormat = void 0;
const label_file_types_1 = require("./label-file-types");
const label_files_shared_1 = require("./label-files-shared");
/**
 * Check whether a JSON object adheres to the schema for a given dataset format
 * @param format Dataset format to check
 * @param json JSON instance
 * @returns valid: true if JSON matches format; if not, returns why not
 */
function checkJsonMatchesFormat(format, json) {
    const valOpts = {
        examineFirstArrayEntryOnly: true,
        examineFirstMapEntryOnly: true
    };
    return validateJsonSchema(format.fn.schema, json, valOpts);
}
exports.checkJsonMatchesFormat = checkJsonMatchesFormat;
/**
 * Check whether a JSON object adheres to a given schema
 * @param schema Schema to validate against
 * @param instance JSON instance
 * @param opts Options
 * @returns valid: true if JSON matches schema; if not, returns why not
 */
function validateJsonSchema(schema, instance, opts) {
    const validateObject = (io, constraint, scope) => {
        if (!constraint) {
            return {
                valid: true
            };
        }
        // If type 'any', do no further validation
        if (constraint.type === 'any') {
            return {
                valid: true
            };
        }
        // If type 'either', check we have any valid match
        if (constraint.type === 'either') {
            for (const possibleConstraint of constraint.possibleTypes) {
                const isValid = validateObject(io, possibleConstraint, scope);
                if (isValid) {
                    return {
                        valid: true
                    };
                }
            }
            // No valid matches
            return {
                valid: false,
                reason: 'Value did not match any valid constraint',
                scope: scope
            };
        }
        // Is the root the correct type?
        const type = Array.isArray(io) ? 'array' : typeof io;
        if (type !== constraint.type) {
            return {
                valid: false,
                reason: `Incorrect type. Got: ${type}, expected: ${constraint.type}`,
                scope: scope
            };
        }
        if (constraint.type === 'object' && !constraint.isMap) {
            // Validate object has required keys & values
            const rootObj = io;
            // For each required key/value:
            for (const [key, value] of Object.entries(constraint.value || [])) {
                // Examine constraint on the current value
                const rootObjEntries = Object.entries(rootObj || []);
                const entry = rootObjEntries.find(([k1, _]) => k1 === key);
                // Value is required but is missing in target object
                if (value.required && typeof entry === 'undefined') {
                    return {
                        valid: false,
                        reason: `Missing value. Expected value for key: ${key}`,
                        scope: scope
                    };
                }
                // Recursively validate this value
                if (entry) {
                    const entryIsValid = validateObject(entry[1], value, [...scope, key]);
                    if (!entryIsValid.valid) {
                        return entryIsValid;
                    }
                }
            }
            // Check we have no additional (unexpected) keys when disallowed
            if (!constraint.allowAllKeys) {
                for (const key of Object.keys(rootObj)) {
                    if (!Object.keys(constraint.value || {}).includes(key)) {
                        return {
                            valid: false,
                            reason: `Got unexpected key: ${key}`,
                            scope: scope
                        };
                    }
                }
            }
        }
        else if (constraint.type === 'object' && constraint.isMap) {
            // Like an 'object', only the keys can be any string, we just enforce types of values
            const rootObj = io;
            // Recursively validate all values
            for (const [key, value] of Object.entries(rootObj)) {
                const entryIsValid = validateObject(value, constraint.values, [...scope, key]);
                if (!entryIsValid.valid) {
                    return entryIsValid;
                }
            }
        }
        else if (constraint.type === 'array') {
            // Validate array has correct types for each entry
            const rootArr = io;
            // Validate array length
            if (typeof constraint.minLength !== 'undefined' && rootArr.length < constraint.minLength) {
                return {
                    valid: false,
                    reason: `Array contains fewer elements than expected. Got: ${rootArr.length}, expected: ${constraint.minLength}`,
                    scope: scope
                };
            }
            if (typeof constraint.maxLength !== 'undefined' && rootArr.length > constraint.maxLength) {
                return {
                    valid: false,
                    reason: `Array contains more elements than expected. Got: ${rootArr.length}, expected: ${constraint.maxLength}`,
                    scope: scope
                };
            }
            // Validate type of each entry
            for (let idx = 0; idx < rootArr.length; idx++) {
                const entryIsValid = validateObject(rootArr[idx], (typeof constraint.values === 'function') ? constraint.values(rootArr[idx]) : constraint.values, [...scope, idx]);
                if (!entryIsValid.valid) {
                    return entryIsValid;
                }
                // Optionally, only validate the first entry
                if (opts && opts.examineFirstArrayEntryOnly) {
                    break;
                }
            }
        }
        else {
            // Primitive type
            // If we have a constraint on the accepted values here, validate them
            if (constraint.valid) {
                const validEntries = constraint.valid.map(cv => JSON.stringify(cv));
                if (validEntries.indexOf(JSON.stringify(io)) === -1) {
                    return {
                        valid: false,
                        reason: `Invalid value. Expected value in: ${JSON.stringify(constraint.valid)}, got: ${JSON.stringify(io)}`,
                        scope: scope
                    };
                }
            }
        }
        return {
            valid: true
        };
    };
    return validateObject(instance, schema, []);
}
exports.validateJsonSchema = validateJsonSchema;
/**
 * Check if a given JSON object matches any known schema from a list of supported types
 * @param jsonFile JSON object to analyze
 * @param formatsToTry List of formats to try; if not given, will try all formats
 * @returns Format info if a match is found
 */
function checkJsonMatchesAnyFormat(jsonFile, formatsToTry) {
    const json = JSON.parse(jsonFile);
    const allFormats = formatsToTry ? formatsToTry : label_file_types_1.supportedJsonLabelFormats;
    for (const format of allFormats) {
        const formatMatch = checkJsonMatchesFormat(format, json);
        if (formatMatch.valid) {
            return format.info;
        }
    }
    return undefined;
}
exports.checkJsonMatchesAnyFormat = checkJsonMatchesAnyFormat;
/**
 * Parse a JSON labels file, converting it to a format usable in EI
 * @param jsonFile JSON file to convert
 * @param format Format to convert from
 * @returns Labels or bounding boxes from the JSON object
 */
function parseJsonLabelsFile(jsonFile, format) {
    const json = JSON.parse(jsonFile);
    const jsonFormat = label_file_types_1.supportedJsonLabelFormats.find(f => f.info.key === format);
    if (!jsonFormat) {
        return {
            success: false,
            reason: `No JSON format matches type '${format}'`
        };
    }
    try {
        // Validate against schema
        const formatMatch = checkJsonMatchesFormat(jsonFormat, json);
        if (!formatMatch.valid) {
            throw new Error(formatMatch.reason);
        }
        // Convert to bounding boxes
        const res = jsonFormat.fn.conversionFunction(json);
        if (res.type === 'object-detection') {
            return {
                success: true,
                match: jsonFormat.info,
                type: 'object-detection',
                labels: res.bboxes,
            };
        }
        else {
            return {
                success: true,
                match: jsonFormat.info,
                type: 'single-label',
                labels: res.labels,
            };
        }
    }
    catch (ex) {
        return {
            success: false,
            reason: `Error parsing type '${jsonFormat.info.name}'; ${ex.message}`
        };
    }
}
exports.parseJsonLabelsFile = parseJsonLabelsFile;
/**
 * Parse a single TXT labels file (e.g. YOLO)
 * @param file TXT file contents
 * @param filename File name; used to work out which sample this label belongs to
 * @param format Dataset format
 * @returns Bounding boxes from the file
 */
function parseTxtLabelsFile(file, filename, format) {
    const labelFormat = label_file_types_1.supportedTxtLabelFormats.find(f => f.info.key === format);
    if (!labelFormat || labelFormat.fn.fileType !== 'txt') {
        return {
            success: false,
            reason: `No txt format matches type '${format}'`
        };
    }
    try {
        const labelsRes = labelFormat.fn.conversionFunction(file);
        let labels = {};
        labels[label_files_shared_1.removeExtension(filename)] = labelsRes;
        return {
            success: true,
            match: labelFormat.info,
            type: 'object-detection',
            labels: labels,
        };
    }
    catch (ex) {
        return {
            success: false,
            reason: `Error parsing type '${labelFormat.info.name}'; ${ex.message}`
        };
    }
}
exports.parseTxtLabelsFile = parseTxtLabelsFile;
//# sourceMappingURL=parse-label-json.js.map